# Stochastic Newton
# input :
#   X      : data
#   lambda : initialization of the Hessian (by default identity)
#   init   : initialization of the model parameter
#   samp   : precision grid for plotting, by default n (only the last iterate)
#   cbeta  : truncation constant c_\beta /k^\beta by default =0 , no need in practice of it
#   beta   : for cbeta
#   f      : by default =0
# output :
#   theta  : complete list of theta



SN=function(X,lambda=diag(ncol(X)-1),init=rep(0,ncol(X)-1),
            samp=c(nrow(X)),f=0,cbeta=0,beta=0.1)
{
  theta=init
  thetalist=c()
  H=lambda
  Y=X[,1]
  X=X[,-1]
  for (i in 1:nrow(X))
  {
    Z=rnorm(ncol(X))
    phi=(X[i,])
    theta=theta +i/(i+f)* (Y[i] - sum(X[i,]*theta))*H%*%phi
    H=H-as.numeric((1+t(phi)%*%H%*%phi))^(-1)*(H%*%(phi))%*%(t(phi)%*%H)
    H=H-as.numeric((1+cbeta*i^(-beta)*t(Z)%*%H%*%Z)^(-1))*cbeta*i^(-beta)*(H%*%Z)%*%(t(Z)%*%H)
    if(length(which(samp==i))==1)
    {
      thetalist=cbind(thetalist,theta)
    }
  }
  list(theta=thetalist)
}

# Weighted  Averaged Stochastic Newton
# input :
#   lambda : init of Hessian (by default identity)
#   c      : constant in gamma_n (step)
#   f      : constant in gamma_n (step)
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   cbeta  : truncation constant c_\beta /k^\beta by default =0 , no need in practice of it
#   beta   : for cbeta
#   alpha  : power in gamma_n (step)
#   b      : power in the log for iterate weigths
# output :
#   theta  : complete list of theta
#   thetaw : list of averaged weighted theta



WASN=function(X,lambda=diag(ncol(X)-1),f=0,c=1,
              init=rep(0,ncol(X)-1),b=2,alpha=0.75,
              samp=c(nrow(X)),cbeta=0,beta=alpha-0.5)
{
  theta=init
  thetaw=init
  thetalist=c()
  thetalistw=c()
  H=lambda
  Y=X[,1]
  X=X[,-1]
  w=0
  for (i in 1:nrow(X))
  {
    Z=rnorm(ncol(X))
    phi=(X[i,])
    theta=theta + c*i*(i+f)^(-alpha)*(Y[i] - sum(X[i,]*theta))*H%*%phi
    w=w+log(i+1)^b
    thetaw=(1 -(log(i+1)^b)/(w))*thetaw + (log(i+1)^b)/(w)*theta
    H=H-as.numeric((1+t(phi)%*%H%*%phi))^(-1)*(H%*%(phi))%*%(t(phi)%*%H)
    H=H-as.numeric((1+cbeta*i^(-beta)*t(Z)%*%H%*%Z))^(-1)*cbeta*i^(-beta)*(H%*%Z)%*%(t(Z)%*%H)
    if(length(which(samp==i))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistw=cbind(thetalistw,thetaw)
    }
  }
  list(theta=thetalist,thetaw=thetalistw)
}

# Averaged Stochastic Gradient
# input :
#   X      : data
#   c      : constant in gamma_n (step)
#   f      : constant in gamma_n (step)
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   alpha  : power in gamma_n (step)
# output :
#   theta  : complete list of theta
#   thetabar : list of standard averaged theta


ASG=function(X,init=rep(0,ncol(X)-1),
             samp=c(nrow(X)),c=1,alpha=0.75,f=0)
{
  theta=init
  thetabar=theta
  thetalist=c()
  thetalistbar=c()
  Y=X[,1]
  X=X[,-1]
  w=0
  for (i in 1:nrow(X))
  {
    Z=rnorm(ncol(X))
    phi=(X[i,])
    theta=theta + c*(i+f)^(-alpha)*(Y[i] - sum(X[i,]*theta))*X[i,]
    thetabar=(i/(i+1))*thetabar + (1/(i+1))*theta
    if(length(which(samp==i))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistbar=cbind(thetalistbar,thetabar)
    }
  }
  list(theta=thetalist,thetabar=thetalistbar)
}


# Adagrad
# input :
#   lambda : init of Hessian (by default identity)
#   c      : constant in step
#   f      : constant in step
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   alpha  : power in gamma_n (step)
# output :
#   theta  : complete list of theta
#   thetabar : list of averaged theta


adagrad=function(X,init=rep(0,ncol(X)-1),
                 samp=c(nrow(X)),c=1,alpha=0.75,f=1)
{
  theta=init
  thetabar=theta
  thetalist=c()
  thetalistbar=c()
  G=rep(f,length(theta))
  Y=X[,1]
  X=X[,-1]
  for (i in 1:nrow(X))
  {
    Z=rnorm(ncol(X))
    phi=(X[i,])
    theta=theta + (sqrt(G))^(-1)*(Y[i] - sum(X[i,]*theta))*X[i,]
    G=G+((Y[i] - sum(X[i,]*theta))*X[i,])^2
    thetabar=(i/(i+1))*thetabar + (1/(i+1))*theta
    if(length(which(samp==i))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistbar=cbind(thetalistbar,thetabar)
    }
  }
  list(theta=thetalist,thetabar=thetalistbar)
}



#### generate sample
n=5000
samp=(1:(n))

#initializing lists for plot
thet0=rep(0,length(samp))
thetg=rep(0,length(samp))
thetg1=rep(0,length(samp))
thetg2=rep(0,length(samp))
thetgav=rep(0,length(samp))
thetgav1=rep(0,length(samp))
thetgav2=rep(0,length(samp))
thetad=rep(0,length(samp))
thetad1=rep(0,length(samp))
thetad2=rep(0,length(samp))
thetadav=rep(0,length(samp))
thetadav1=rep(0,length(samp))
thetadav2=rep(0,length(samp))
thetg=rep(0,length(samp))
thetg1=rep(0,length(samp))
thetg2=rep(0,length(samp))
thetgav=rep(0,length(samp))
thetgav1=rep(0,length(samp))
thetgav2=rep(0,length(samp))
thet=rep(0,length(samp))
thetbar=rep(0,length(samp))
thetw=rep(0,length(samp))
thet01=rep(0,length(samp))
thet1=rep(0,length(samp))
thetbar1=rep(0,length(samp))
thetw1=rep(0,length(samp))
thet02=rep(0,length(samp))
thet2=rep(0,length(samp))
thetbar2=rep(0,length(samp))
thetw2=rep(0,length(samp))



#generate sample
d=10
theta=1:d-d/2
a=matrix(rnorm(d^2),nrow=d)
mat=svd(a%*%t(a))$u
mat=diag(d)

gen=function(n,theta,sigma=1,mat)
{
  d=length(theta)
  Y=c()
  X=c()
  for (i in 1:n)
  {
    Z=mat%*%((1:d)/d*rnorm(d))
    Y=c(Y,sum(theta*Z) + rnorm(1,0,sigma))
    X=rbind(X,t(Z))
  }
  cbind(Y,X)
}

# number of samples

nech=50
f=100

# to calculate square errors
mattheta=matrix(rep(theta,length(samp)),nrow=length(theta))

for (k in 1:nech)
{
  init0=theta
  z=rnorm(length(theta))
  z=z/sqrt(sum(z^2))
  init=init0+z
  init1=init0+2*z
  init2=init0+5*z
  X=gen(n,theta=theta,mat=mat)
  sn=SN(X,init=init,samp=samp,f=f)
  asn=WASN(X,init=init,samp=samp,b=0,f=f)
  ad=adagrad(X,init=init,samp=samp,f=f)
  ad1=adagrad(X,init=init1,samp=samp,f=f)
  ad2=adagrad(X,init=init2,samp=samp,f=f)
  gr=ASG(X,init=init,samp=samp,f=f)
  gr1=ASG(X,init=init1,samp=samp,f=f)
  gr2=ASG(X,init=init2,samp=samp,f=f)
  wasn=WASN(X,init=init,samp=samp,b=2,f=f)
  sn1=SN(X,init=init1,samp=samp,f=f)
  asn1=WASN(X,init=init1,samp=samp,b=0,f=f)
  wasn1=WASN(X,init=init1,samp=samp,b=2,f=f)
  sn2=SN(X,init=init2,samp=samp,f=f)
  asn2=WASN(X,init=init2,samp=samp,b=0,f=f)
  wasn2=WASN(X,init=init2,samp=samp,b=2,f=f)
  thet0=thet0+colSums((mattheta - asn$theta)^2)
  thet=thet+colSums((mattheta - sn$theta)^2)
  thetg=thetg+colSums((mattheta - gr$theta)^2)
  thetg1=thetg1+colSums((mattheta - gr1$theta)^2)
  thetg2=thetg2+colSums((mattheta - gr2$theta)^2)
  thetgav=thetgav+colSums((mattheta - gr$thetabar)^2)
  thetgav1=thetgav1+colSums((mattheta - gr1$thetabar)^2)
  thetgav2=thetgav2+colSums((mattheta - gr2$thetabar)^2)
  thetad=thetad+colSums((mattheta - ad$theta)^2)
  thetad1=thetad1+colSums((mattheta - ad1$theta)^2)
  thetad2=thetad2+colSums((mattheta - ad2$theta)^2)
  thetadav=thetadav+colSums((mattheta - ad$thetabar)^2)
  thetadav1=thetadav1+colSums((mattheta - ad1$thetabar)^2)
  thetadav2=thetadav2+colSums((mattheta - ad2$thetabar)^2)
  thetbar=thetbar+colSums((mattheta - asn$thetaw)^2)
  thetw=thetw+colSums((mattheta - wasn$thetaw)^2)
  thet01=thet01+colSums((mattheta - asn1$theta)^2)
  thet1=thet1+colSums((mattheta - sn1$theta)^2)
  thetbar1=thetbar1+colSums((mattheta - asn1$thetaw)^2)
  thetw1=thetw1+colSums((mattheta - wasn1$thetaw)^2)
  thet02=thet02+colSums((mattheta - asn2$theta)^2)
  thet2=thet2+colSums((mattheta - sn2$theta)^2)
  thetbar2=thetbar2+colSums((mattheta - asn2$thetaw)^2)
  thetw2=thetw2+colSums((mattheta - wasn2$thetaw)^2)
}
thet0=thet0/nech
thet=thet/nech
thetad=thetad/nech
thetad1=thetad1/nech
thetad2=thetad2/nech
thetadav=thetadav/nech
thetadav1=thetadav1/nech
thetadav2=thetadav2/nech
thetg=thetg/nech
thetg1=thetg1/nech
thetg2=thetg2/nech
thetgav=thetgav/nech
thetgav1=thetgav1/nech
thetgav2=thetgav2/nech
thetbar=thetbar/nech
thetw=thetw/nech
thet01=thet01/nech
thet1=thet1/nech
thetbar1=thetbar1/nech
thetw1=thetw1/nech
thet02=thet02/nech
thet2=thet2/nech
thetbar2=thetbar2/nech
thetw2=thetw2/nech

library(latex2exp)

par(mfrow=c(1,3),oma=c(4,3,0,0),mar=c(1,1,1,1))
matplot(samp,cbind(thet,thetbar,thetg,thet0,thetgav,thetw,thetad,thetadav),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error", main="",
        lty=c(2,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),
                  TeX("SN with step (n+100)^{-1}"),"ASN (standard)",
                  "ASN (log)","SGD","ASGD","Adagrad","Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))
matplot(samp,cbind(thet1,thetbar1,thetg1,thet01,thetgav1,thetw1,thetad1,thetadav1),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error",main="",
        lty=c(1,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),
                  TeX("SN with step (n+100)^{-1}"),"ASN (standard)",
                  "ASN (log)","SGD","ASGD","Adagrad","Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))
matplot(samp,cbind(thet2,thetbar2,thetg2,thet02,thetgav2,thetw2,thetad2,thetadav2),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error",
        main="",lty=c(2,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),TeX("SN with step (n+100)^{-1}"),
                  "ASN (standard)","ASN (log)","SGD","ASGD","Adagrad",
                  "Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))

