# Indicator function

ind <- function(x,k) ifelse(x %in% k, 1,0)



# Stochastic Newton
# input :
#   X      : data
#   lambda : initialization of the Hessian (by default identity)
#   init   : initialization of the model parameter
#   samp   : precision grid for plotting, by default n (only the last iterate)
#   cbeta  : truncation constant c_\beta /k^\beta by default =0 , no need in practice of it
#   beta   : for cbeta
#   f      : by default =0
# output :
#   theta  : complete list of theta






SN=function(X,f=0,lambda=diag(ncol(X)-1),
            init=rep(0,ncol(X)-1),samp=c(nrow(X)),
            cbeta=0,beta=alpha-0.5)
{
  theta=init
  thetalist=c()
  Q=lambda
  Y=X[,1]
  X=X[,2:ncol(X)]
  for (j in 1:nrow(X) ) {
    y       = Y[j]
    phi    = c( X[j,])
    pix    = exp(sum(phi*theta)) / (1 + exp(sum(phi*theta)) )
    a_n     = sqrt(pix * (1 - pix))
    a_n2     = max(c(sqrt(pix * (1 - pix)),j^-beta))
    Phi    = a_n2 * c( X[j,])
    fn      = as.numeric( t(Phi) %*% Q %*% Phi )
    theta = theta + Q %*% phi * (y - pix)*j/(j+f)  
    Q      = Q - (Q %*% Phi %*% t(Phi) %*% Q) / (1 + fn) 
    if(length(which(samp==j))==1)
    {
      thetalist=cbind(thetalist,theta)
    }
  }
  list(theta=thetalist)
}



# Averaged Stochastic Gradient
# input :
#   X      : data
#   c      : constant in gamma_n (step)
#   f      : constant in gamma_n (step)
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   alpha  : power in gamma_n (step)
# output :
#   theta  : complete list of theta
#   thetabar : list of standard averaged theta


ASG=function(X,lambda=diag(ncol(X)-1),f=0,c=1,
             init=rep(0,ncol(X)-1),alpha=0.75,
             samp=c(nrow(X)),cbeta=0,beta=alpha-0.5)
{
  theta=init
  thetabar=theta
  thetalist=c()
  thetalistbar=c()
  Y=X[,1]
  X=X[,2:ncol(X)]
  for (j in 1:nrow(X) ) {
    y       = Y[j]
    phi    = c(X[j,])
    pix    = exp(sum(phi*theta)) / (1 + exp(sum(phi*theta)) )
    a_n     = sqrt(pix * (1 - pix))
    a_n2     = max(c(sqrt(pix * (1 - pix)),j^-beta))
    Phi    = a_n2 * c(X[j,])
    fn      = as.numeric( 0.25*t(phi) %*% Q %*% phi )
    theta = theta + c*(j+f)^(-alpha)*  phi * (y - pix)
    thetabar=((j)/(j+1))* thetabar + 1/(j+1)*theta
    if(length(which(samp==j))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistbar=cbind(thetalistbar,thetabar)
    }
  }
  list(theta=thetalist,thetabar=thetalistbar)
}


# Weighted  Averaged Stochastic Newton
# input :
#   lambda : init of Hessian (by default identity)
#   c      : constant in gamma_n (step)
#   f      : constant in gamma_n (step)
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   cbeta  : truncation constant c_\beta /k^\beta by default =0 , no need in practice of it
#   beta   : for cbeta
#   alpha  : power in gamma_n (step)
#   b      : power in the log for iterate weigths
# output :
#   theta  : complete list of theta
#   thetaw : list of averaged weighted theta

WASN=function(X,lambda=diag(ncol(X)-1),f=0,c=1,
             init=rep(0,ncol(X)-1),b=2,alpha=0.75,
             samp=c(nrow(X)),cbeta=0,beta=alpha-0.5)
{
  theta=init
  thetabar=theta
  thetalist=c()
  thetalistbar=c()
  Q=lambda
  Y=X[,1]
  X=X[,2:ncol(X)]
  w=0
  for (j in 1:nrow(X) ) {
    y       = Y[j]
    phi    = c( X[j,])
    pix    = exp(sum(phi*theta)) / (1 + exp(sum(phi*theta)) )
    pixbar    = exp(sum(phi*thetabar)) / (1 + exp(sum(phi*thetabar)) )
    a_n     = sqrt(pixbar * (1 - pixbar))
    a_n2     = max(c(sqrt(pixbar * (1 - pixbar)),j^-beta))
    Phi    = a_n2 * c( X[j,])
    fn      = as.numeric( t(Phi) %*% Q %*% Phi )
    theta = theta + c*j/(j+f)^(alpha)* Q %*% phi * (y - pix)
    w=w+log(j+1)^b
    thetabar=(1 -(log(j+1)^b)/(w))*thetabar + (log(j+1)^b)/(w)*theta
    Q      = Q - (Q %*% Phi %*% t(Phi) %*% Q) / (1 + fn) 
    if(length(which(samp==j))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistbar=cbind(thetalistbar,thetabar)
    }
  }
  list(theta=thetalist,thetaw=thetalistbar)
}


# Adagrad
# input :
#   lambda : init of Hessian (by default identity)
#   c      : constant in gamma_n (step)
#   f      : constant in gamma_n (step)
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   alpha  : power in gamma_n (step)
# output :
#   theta  : complete list of theta
#   thetabar : list of averaged theta

adagrad=function(X,f=1,init=rep(0,ncol(X)-1),
                 c=1,alpha=0.75,samp=c(nrow(X)))
{
  theta=init
  thetabar=theta
  thetalist=c()
  thetalistbar=c()
  Y=X[,1]
  X=X[,2:ncol(X)]
  G=rep(f,length(theta))
  for (j in 1:nrow(X) ) {
    y       = Y[j]
    phi    = c( X[j,])
    pix    = exp(sum(phi*theta)) / (1 + exp(sum(phi*theta)) )
    a_n     = sqrt(pix * (1 - pix))
    a_n2     = max(c(sqrt(pix * (1 - pix)),j^-beta))
    Phi    = a_n2 * c(X[j,])
    fn      = as.numeric( 0.25*t(phi) %*% Q %*% phi )
    theta = theta + c*G^(-1/2)*  phi * (y - pix)
    G=G+phi^2*(y-pix)^2
    thetabar=((j)/(j+1))* thetabar + 1/(j+1)*theta
    if(length(which(samp==j))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistbar=cbind(thetalistbar,thetabar)
    }
  }
  list(theta=thetalist,thetabar=thetalistbar)
}







#sample size
n=5000
samp=1*(1:(n/1))



#initializing lists for plot
thet0=rep(0,length(samp))
thetg=rep(0,length(samp))
thetg1=rep(0,length(samp))
thetg2=rep(0,length(samp))
thetgav=rep(0,length(samp))
thetgav1=rep(0,length(samp))
thetgav2=rep(0,length(samp))
thetad=rep(0,length(samp))
thetad1=rep(0,length(samp))
thetad2=rep(0,length(samp))
thetadav=rep(0,length(samp))
thetadav1=rep(0,length(samp))
thetadav2=rep(0,length(samp))
thetg=rep(0,length(samp))
thetg1=rep(0,length(samp))
thetg2=rep(0,length(samp))
thetgav=rep(0,length(samp))
thetgav1=rep(0,length(samp))
thetgav2=rep(0,length(samp))
thet=rep(0,length(samp))
thetbar=rep(0,length(samp))
thetw=rep(0,length(samp))
thet01=rep(0,length(samp))
thet1=rep(0,length(samp))
thetbar1=rep(0,length(samp))
thetw1=rep(0,length(samp))
thet02=rep(0,length(samp))
thet2=rep(0,length(samp))
thetbar2=rep(0,length(samp))
thetw2=rep(0,length(samp))



#generate sample
theta=rep(1,10)
n=5000
f=100
d=length(theta)
a=matrix(rnorm(d^2),nrow=d)
mat=svd(a%*%t(a))$u

gen=function(theta,n,mat)
{
  Y=c()
  X=c()
  d=length(theta)
  for (i in 1:n)
  {
    X2 = mat%*%((1:d)/d*rnorm(d))
    Z   = sum(X2*theta)
    X=rbind(X,t(X2))
    Pix = exp(Z) / (1 + exp(Z))
    U = runif(1)
    Y2 = (U < Pix) * 1 + (U >= Pix) * 0
    Y=c(Y,Y2)
  }
  return(cbind(Y,X))
}



# number of samples

nech=50
f=100

# to calculate square errors
mattheta=matrix(rep(theta,length(samp)),nrow=length(theta))

for (k in 1:nech)
{
  init0=theta
  z=rnorm(length(theta))
  z=z/sqrt(sum(z^2))
  init=init0+z
  init1=init0+2*z
  init2=init0+5*z
  X=gen(n,theta=theta,mat=mat)
  sn=SN(X,init=init,samp=samp,f=f)
  asn=WASN(X,init=init,samp=samp,b=0,f=f)
  ad=adagrad(X,init=init,samp=samp,f=f)
  ad1=adagrad(X,init=init1,samp=samp,f=f)
  ad2=adagrad(X,init=init2,samp=samp,f=f)
  gr=ASG(X,init=init,samp=samp,f=f)
  gr1=ASG(X,init=init1,samp=samp,f=f)
  gr2=ASG(X,init=init2,samp=samp,f=f)
  wasn=WASN(X,init=init,samp=samp,b=2,f=f)
  sn1=SN(X,init=init1,samp=samp,f=f)
  asn1=WASN(X,init=init1,samp=samp,b=0,f=f)
  wasn1=WASN(X,init=init1,samp=samp,b=2,f=f)
  sn2=SN(X,init=init2,samp=samp,f=f)
  asn2=WASN(X,init=init2,samp=samp,b=0,f=f)
  wasn2=WASN(X,init=init2,samp=samp,b=2,f=f)
  thet0=thet0+colSums((mattheta - asn$theta)^2)
  thet=thet+colSums((mattheta - sn$theta)^2)
  thetg=thetg+colSums((mattheta - gr$theta)^2)
  thetg1=thetg1+colSums((mattheta - gr1$theta)^2)
  thetg2=thetg2+colSums((mattheta - gr2$theta)^2)
  thetgav=thetgav+colSums((mattheta - gr$thetabar)^2)
  thetgav1=thetgav1+colSums((mattheta - gr1$thetabar)^2)
  thetgav2=thetgav2+colSums((mattheta - gr2$thetabar)^2)
  thetad=thetad+colSums((mattheta - ad$theta)^2)
  thetad1=thetad1+colSums((mattheta - ad1$theta)^2)
  thetad2=thetad2+colSums((mattheta - ad2$theta)^2)
  thetadav=thetadav+colSums((mattheta - ad$thetabar)^2)
  thetadav1=thetadav1+colSums((mattheta - ad1$thetabar)^2)
  thetadav2=thetadav2+colSums((mattheta - ad2$thetabar)^2)
  thetbar=thetbar+colSums((mattheta - asn$thetaw)^2)
  thetw=thetw+colSums((mattheta - wasn$thetaw)^2)
  thet01=thet01+colSums((mattheta - asn1$theta)^2)
  thet1=thet1+colSums((mattheta - sn1$theta)^2)
  thetbar1=thetbar1+colSums((mattheta - asn1$thetaw)^2)
  thetw1=thetw1+colSums((mattheta - wasn1$thetaw)^2)
  thet02=thet02+colSums((mattheta - asn2$theta)^2)
  thet2=thet2+colSums((mattheta - sn2$theta)^2)
  thetbar2=thetbar2+colSums((mattheta - asn2$thetaw)^2)
  thetw2=thetw2+colSums((mattheta - wasn2$thetaw)^2)
}
thet0=thet0/nech
thet=thet/nech
thetad=thetad/nech
thetad1=thetad1/nech
thetad2=thetad2/nech
thetadav=thetadav/nech
thetadav1=thetadav1/nech
thetadav2=thetadav2/nech
thetg=thetg/nech
thetg1=thetg1/nech
thetg2=thetg2/nech
thetgav=thetgav/nech
thetgav1=thetgav1/nech
thetgav2=thetgav2/nech
thetbar=thetbar/nech
thetw=thetw/nech
thet01=thet01/nech
thet1=thet1/nech
thetbar1=thetbar1/nech
thetw1=thetw1/nech
thet02=thet02/nech
thet2=thet2/nech
thetbar2=thetbar2/nech
thetw2=thetw2/nech

library(latex2exp)

par(mfrow=c(1,3),oma=c(4,3,0,0),mar=c(1,1,1,1))
matplot(samp,cbind(thet,thetbar,thetg,thet0,thetgav,thetw,thetad,thetadav),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error", main="",
        lty=c(2,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),
                  TeX("SN with step (n+100)^{-1}"),"ASN (standard)",
                  "ASN (log)","SGD","ASGD","Adagrad","Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))
matplot(samp,cbind(thet1,thetbar1,thetg1,thet01,thetgav1,thetw1,thetad1,thetadav1),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error",main="",
        lty=c(1,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),
                  TeX("SN with step (n+100)^{-1}"),"ASN (standard)",
                  "ASN (log)","SGD","ASGD","Adagrad","Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))
matplot(samp,cbind(thet2,thetbar2,thetg2,thet02,thetgav2,thetw2,thetad2,thetadav2),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error",
        main="",lty=c(2,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),TeX("SN with step (n+100)^{-1}"),
                  "ASN (standard)","ASN (log)","SGD","ASGD","Adagrad",
                  "Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))
