# Indicator function

ind <- function(x,k) ifelse(x %in% k, 1,0)



# Stochastic Newton
# input :
#   X      : data
#   lambda : initialization of the Hessian (by default identity)
#   K      : number of classes
#   init   : initialization of the model parameter
#   samp   : precision grid for plotting, by default n (only the last iterate)
#   cbeta  : truncation constant c_\beta /k^\beta by default =0 , no need in practice of it
#   beta   : for cbeta
#   f      : by default =0
# output :
#   theta  : complete list of theta
SN=function(X,lambda=diag((ncol(X)-1)*K),K=10,
            init=rep(0,(ncol(X)-1)*K),samp=c(nrow(X)),
            cbeta=0,beta=0.1,f=0)
{
  theta=init
  thetalist=c()
  H=lambda
  H2=lambda^(-1)
  for (i in 1:nrow(X))
  {
    Z=rnorm((ncol(X)-1)*K)
    phi=X[i,2:ncol(X)]
    p=exp(matrix(theta,byrow=T,ncol=ncol(X)-1)%*%c(phi)-max(matrix(theta,byrow=T,ncol=ncol(X)-1)%*%c(phi)))
    p=p/sum(p)
    grad=c()
    for (j in 1:K)
    {
      grad=c(grad,phi*(p[j]-ind(X[i,1],j) ))
    }
    theta=theta - H%*%grad*i/(i+f)
    H=H-as.numeric((1+t(grad)%*%H%*%grad)^(-1))*(H%*%(grad))%*%(t(grad)%*%H)
    H=H-as.numeric((1+cbeta*i^(-beta)*t(Z)%*%H%*%Z)^(-1))*cbeta*i^(-beta)*(H%*%Z)%*%(t(Z)%*%H)
    if(length(which(samp==i))==1)
    {
      thetalist=cbind(thetalist,theta)
    }
  }
  list(theta=thetalist,H=H)
}


# Averaged Stochastic Gradient
# input :
#   X      : data
#   c      : constant in gamma_n (step)
#   f      : constant in gamma_n (step)
#   K      : nb of classes
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   alpha  : power in gamma_n (step)
# output :
#   theta  : complete list of theta
#   thetabar : list of standard averaged theta
ASG=function(X,c=1,K=10,init=rep(0,(ncol(X)-1)*K),
             samp=c(nrow(X)),alpha=0.75,f=0)
{
  theta=init
  thetabar=theta
  thetalist=c()
  thetalistbar=c()
  for (i in 1:nrow(X))
  {
    phi=X[i,2:ncol(X)]
    p=exp(matrix(theta,byrow=T,ncol=ncol(X)-1)%*%(phi)-max(matrix(theta,byrow=T,ncol=ncol(X)-1)%*%(phi)))
    p=p/sum(p)
    pbar=exp(matrix(thetabar,byrow=T,ncol=ncol(X)-1)%*%(phi)-max(matrix(thetabar,byrow=T,ncol=ncol(X)-1)%*%(phi)))
    pbar=pbar/sum(pbar)
    grad=c()
    gradbar=c()
    for (j in 1:K)
    {
      grad=c(grad,phi*(p[j]-ind(X[i,1],j) ))
    }
    theta=theta - c/(i+f)^(alpha)*grad
    thetabar=(i/(i+1))*thetabar + (1/(i+1))*theta
    if(length(which(samp==i))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistbar=cbind(thetalistbar,thetabar)
    }
  }
  list(theta=thetalist,thetabar=thetalistbar)
}


# Weighted  Averaged Stochastic Newton
# input :
#   lambda : init of Hessian (by default identity)
#   c      : constant in gamma_n (step)
#   f      : constant in gamma_n (step)
#   K      : nb of classes
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   cbeta  : truncation constant c_\beta /k^\beta by default =0 , no need in practice of it
#   beta   : for cbeta
#   alpha  : power in gamma_n (step)
#   b      : power in the log for iterate weigths
# output :
#   theta  : complete list of theta
#   thetaw : list of averaged weighted theta


WASN=function(X,lambda=diag((ncol(X)-1)*K),c=1,K=10,
              init=rep(0,(ncol(X)-1)*K),samp=c(nrow(X)),
              cbeta=0,beta=0.1,alpha=0.75,b=2,f=0)
{
  theta=init
  thetaw=theta
  thetalist=c()
  thetalistw=c()
  H=lambda
  w=0
  for (i in 1:nrow(X))
  {
    Z=rnorm((ncol(X)-1)*K)
    phi=X[i,2:ncol(X)]
    p=exp(matrix(theta,byrow=T,ncol=ncol(X)-1)%*%(phi)-max(matrix(theta,byrow=T,ncol=ncol(X)-1)%*%(phi)))
    p=p/sum(p)
    pbar=exp(matrix(thetaw,byrow=T,ncol=ncol(X)-1)%*%(phi)-max(matrix(thetaw,byrow=T,ncol=ncol(X)-1)%*%(phi)))
    pbar=pbar/sum(pbar)
    grad=c()
    gradbar=c()
    for (j in 1:K)
    {
      grad=c(grad,phi*(p[j]-ind(X[i,1],j) ))
      gradbar=c(gradbar,phi*(pbar[j]-ind(X[i,1],j) ))
    }
    theta=theta - c*i/(i+f)^(alpha)*H%*%grad
    w=w+log(i+1)^b
    thetaw= (1- (log(i+1)^b/w))*thetaw + (log(i+1)^b/w) * theta
    H=H-as.numeric((1+t(gradbar)%*%H%*%gradbar)^(-1))*(H%*%(gradbar))%*%(t(gradbar)%*%H)
    H=H-as.numeric((1+cbeta*i^(-beta)*t(Z)%*%H%*%Z)^(-1))*cbeta*i^(-beta)*(H%*%Z)%*%(t(Z)%*%H)
    if(length(which(samp==i))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistw=cbind(thetalistw,thetaw)
    }
  }
  list(theta=thetalist,thetaw=thetalistw,H=H)
}


# Adagrad
# input :
#   c      : constant in gamma_n (step)
#   f      : constant in gamma_n (step)
#   K      : nb of classes
#   init   : init of the model parameter
#   samp   : by default n (only the last iterate), precision grid for plotting
#   alpha  : power in gamma_n (step)
# output :
#   theta  : complete list of theta
#   thetabar : list of averaged theta


adagrad=function(X,c=1,K=10,init=rep(0,(ncol(X)-1)*K),
                 samp=c(nrow(X)),cbeta=0,beta=0.1,
                 alpha=0.75,f=0)
{
  theta=init
  thetabar=theta
  thetalist=c()
  thetalistw=c()
  thetalistbar=c()
  G=rep(f,length(theta))
  chilist=c()
  for (i in 1:nrow(X))
  {
    Z=rnorm((ncol(X)-1)*K)
    phi=X[i,2:ncol(X)]
    p=exp(matrix(theta,byrow=T,ncol=ncol(X)-1)%*%(phi)-max(matrix(theta,byrow=T,ncol=ncol(X)-1)%*%(phi)))
    p=p/sum(p)
    grad=c()
    for (j in 1:K)
    {
      grad=c(grad,phi*(p[j]-ind(X[i,1],j) ))
    }
    theta=theta - c*G^(-alpha)*grad
    G=G+grad^2
    thetabar=(i/(i+1))*thetabar + (1/(i+1))*theta
    if(length(which(samp==i))==1)
    {
      thetalist=cbind(thetalist,theta)
      thetalistbar=cbind(thetalistbar,thetabar)
    }
  }
  list(theta=thetalist,thetabar=thetalistbar)
}


## Generate a sample

gen=function(n,theta,mat)
{
  d=ncol(theta)
  X=c()
  for (i in 1:n)
  {
    Z=mat%*%((1:d)/d*c(rnorm(d)))
    p=exp(theta%*%Z - max(theta%*%Z))
    p=p/sum(p)
    Y=sample(1:length(p),1,prob=p)
    X=rbind(X,c(Y,Z))
  }
  X
}


# Generate theta

K=3
d=3
Z=rnorm(d*K)
Z=Z/sqrt(sum(Z^2))
theta=matrix(Z,nrow=K)

#generate covariance
a=matrix(rnorm(d^2),nrow=d)
mat=svd(a%*%t(a))$u
mat=diag(d)



#sample size
n=5000
samp=1*(1:(n/1))



#initializing lists for plot
thet0=rep(0,length(samp))
thetg=rep(0,length(samp))
thetg1=rep(0,length(samp))
thetg2=rep(0,length(samp))
thetgav=rep(0,length(samp))
thetgav1=rep(0,length(samp))
thetgav2=rep(0,length(samp))
thetad=rep(0,length(samp))
thetad1=rep(0,length(samp))
thetad2=rep(0,length(samp))
thetadav=rep(0,length(samp))
thetadav1=rep(0,length(samp))
thetadav2=rep(0,length(samp))
thetg=rep(0,length(samp))
thetg1=rep(0,length(samp))
thetg2=rep(0,length(samp))
thetgav=rep(0,length(samp))
thetgav1=rep(0,length(samp))
thetgav2=rep(0,length(samp))
thet=rep(0,length(samp))
thetbar=rep(0,length(samp))
thetw=rep(0,length(samp))
thet01=rep(0,length(samp))
thet1=rep(0,length(samp))
thetbar1=rep(0,length(samp))
thetw1=rep(0,length(samp))
thet02=rep(0,length(samp))
thet2=rep(0,length(samp))
thetbar2=rep(0,length(samp))
thetw2=rep(0,length(samp))



cbeta=0
#beta=0.25

# number of samples

nech=50
f=100

# to calculate square errors
theta2=c(t(theta))
mattheta=matrix(rep(theta2,length(samp)),ncol=length(samp))
for (k in 1:nech)
{
  init0=c(t(theta))
  z=rnorm(length(theta))
  z=z/sqrt(sum(z^2))
  init=init0+z
  init1=init0+2*z
  init2=init0+5*z
  X=gen(n,theta=theta,mat=mat)
  sn=SN(X,K=K,init=init,samp=samp,f=f)
  asn=WASN(X,K=K,init=init,samp=samp,b=0,f=f)
  ad=adagrad(X,K=K,init=init,samp=samp,b=2,f=f)
  ad1=adagrad(X,K=K,init=init1,samp=samp,b=2,f=f)
  ad2=adagrad(X,K=K,init=init2,samp=samp,b=2,f=f)
  gr=ASG(X,K=K,init=init,samp=samp,f=f)
  gr1=ASG(X,K=K,init=init1,samp=samp,f=f)
  gr2=ASG(X,K=K,init=init2,samp=samp,f=f)
  wasn=WASN(X,K=K,init=init,samp=samp,b=2,f=f)
  sn1=SN(X,K=K,init=init1,samp=samp,f=f)
  asn1=WASN(X,K=K,init=init1,samp=samp,b=0,f=f)
  wasn1=WASN(X,K=K,init=init1,samp=samp,b=2,f=f)
  sn2=SN(X,K=K,init=init2,samp=samp,f=f)
  asn2=WASN(X,K=K,init=init2,samp=samp,b=0,f=f)
  wasn2=WASN(X,K=K,init=init2,samp=samp,b=2,f=f)
  thet0=thet0+colSums((mattheta - asn$theta)^2)
  thet=thet+colSums((mattheta - sn$theta)^2)
  thetg=thetg+colSums((mattheta - gr$theta)^2)
  thetg1=thetg1+colSums((mattheta - gr1$theta)^2)
  thetg2=thetg2+colSums((mattheta - gr2$theta)^2)
  thetgav=thetgav+colSums((mattheta - gr$thetabar)^2)
  thetgav1=thetgav1+colSums((mattheta - gr1$thetabar)^2)
  thetgav2=thetgav2+colSums((mattheta - gr2$thetabar)^2)
  thetad=thetad+colSums((mattheta - ad$theta)^2)
  thetad1=thetad1+colSums((mattheta - ad1$theta)^2)
  thetad2=thetad2+colSums((mattheta - ad2$theta)^2)
  thetadav=thetadav+colSums((mattheta - ad$thetabar)^2)
  thetadav1=thetadav1+colSums((mattheta - ad1$thetabar)^2)
  thetadav2=thetadav2+colSums((mattheta - ad2$thetabar)^2)
  thetbar=thetbar+colSums((mattheta - asn$thetaw)^2)
  thetw=thetw+colSums((mattheta - wasn$thetaw)^2)
  thet01=thet01+colSums((mattheta - asn1$theta)^2)
  thet1=thet1+colSums((mattheta - sn1$theta)^2)
  thetbar1=thetbar1+colSums((mattheta - asn1$thetaw)^2)
  thetw1=thetw1+colSums((mattheta - wasn1$thetaw)^2)
  thet02=thet02+colSums((mattheta - asn2$theta)^2)
  thet2=thet2+colSums((mattheta - sn2$theta)^2)
  thetbar2=thetbar2+colSums((mattheta - asn2$thetaw)^2)
  thetw2=thetw2+colSums((mattheta - wasn2$thetaw)^2)
}
thet0=thet0/nech
thet=thet/nech
thetad=thetad/nech
thetad1=thetad1/nech
thetad2=thetad2/nech
thetadav=thetadav/nech
thetadav1=thetadav1/nech
thetadav2=thetadav2/nech
thetg=thetg/nech
thetg1=thetg1/nech
thetg2=thetg2/nech
thetgav=thetgav/nech
thetgav1=thetgav1/nech
thetgav2=thetgav2/nech
thetbar=thetbar/nech
thetw=thetw/nech
thet01=thet01/nech
thet1=thet1/nech
thetbar1=thetbar1/nech
thetw1=thetw1/nech
thet02=thet02/nech
thet2=thet2/nech
thetbar2=thetbar2/nech
thetw2=thetw2/nech

library(latex2exp)

par(mfrow=c(1,3),oma=c(4,3,0,0),mar=c(1,1,1,1))
matplot(samp,cbind(thet,thetbar,thetg,thet0,thetgav,thetw,thetad,thetadav),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error", main="",
        lty=c(2,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),
                  TeX("SN with step (n+100)^{-1}"),"ASN (standard)",
                  "ASN (log)","SGD","ASGD","Adagrad","Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))
matplot(samp,cbind(thet1,thetbar1,thetg1,thet01,thetgav1,thetw1,thetad1,thetadav1),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error",main="",
        lty=c(1,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),
                  TeX("SN with step (n+100)^{-1}"),"ASN (standard)",
                  "ASN (log)","SGD","ASGD","Adagrad","Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))
matplot(samp,cbind(thet2,thetbar2,thetg2,thet02,thetgav2,thetw2,thetad2,thetadav2),
        'l',log='xy',xlab="Sample size",ylab="Quadratic mean error",
        main="",lty=c(2,1,2,1,1,2,1,2),
        col=c("black","blue","red","black","red","blue","orange","orange"))
legend("bottomleft",cex=0.9,
       legend = c(TeX("SN with step (n+100)^{-0.75}"),TeX("SN with step (n+100)^{-1}"),
                  "ASN (standard)","ASN (log)","SGD","ASGD","Adagrad",
                  "Adagrad (averaged)"),
       lty=c(1,2,1,2,1,2,1,2),
       col=c("black","black","blue","blue","red","red","orange","orange"))
